from datetime import datetime, timedelta

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.session import object_session
from sqlalchemy.sql.schema import Column, ColumnDefault
from sqlalchemy.sql.schema import ForeignKey, UniqueConstraint, Index
from sqlalchemy import Boolean, SmallInteger, Integer, BigInteger, Float, String, Text, Unicode, UnicodeText, Binary, DateTime, Enum
from sqlalchemy.orm import relationship, backref, remote, foreign

from sqlalchemy.sql.expression import select, and_, func

Base = declarative_base()

# TODO: server grouping
# TODO: service grouping


class Server(Base):
    """ A server being monitored """
    __tablename__ = 'servers'

    id = Column(Integer, primary_key=True, nullable=False)
    title = Column(Unicode(64), nullable=False, doc="Server title")
    name = Column(String(32), nullable=False, doc="Server machine name (as reported from the remote)")
    key = Column(String(32), nullable=False, doc="Authentication key")
    ip = Column(String(46), nullable=True, doc='IP-address')

    __table_args__ = (
        UniqueConstraint(name),
    )

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.title or self.name

class Service(Base):
    """ A service being monitored """
    __tablename__ = 'services'

    id = Column(Integer, primary_key=True, nullable=False)
    server_id = Column(Integer, ForeignKey(Server.id, ondelete='CASCADE'), nullable=False, doc="Server id")

    period = Column(Integer, nullable=True, doc="Expected reporting period, seconds")
    timed_out = Column(Boolean, nullable=False, default=False, doc="Is currently timed out?")

    name = Column(String(32), nullable=False, doc="Service machine name (as reported from the remote)")
    title = Column(Unicode(64), nullable=False, default=u'', doc="Service title")

    server = relationship(Server, foreign_keys=server_id, backref=backref('services', passive_deletes=True))

    __table_args__ = (
        UniqueConstraint(server_id, name),
    )

    def update_timed_out(self):
        """ Update service's `timed_out` field
        :returns: How long ago it was last seen (if timed out)
        :rtype: timedelta
        """
        if not self.state:
            return timedelta(seconds=0)
        dt = datetime.utcnow() - self.state.rtime
        self.timed_out = dt > timedelta(seconds=self.period)
        return dt

    def __str__(self):
        return self.name

    def __unicode__(self):
        return self.title or self.name


class state_t(int):
    """ Comparable enum for state """
    states = ('UNK', 'OK', 'WARN', 'FAIL')

    UNK = 0
    OK = 1
    WARN = 2
    FAIL = 3

    def __new__(cls, state):
        intval = cls.states.index(state)
        return super(state_t, cls).__new__(cls, intval)

    @classmethod
    def is_valid(cls, state):
        """ Test whether the state value is valid """
        try:
            cls.states.index(state)
            return True
        except ValueError:
            return False



class ServiceState(Base):
    """ Service state """
    __tablename__ = 'service_states'

    id = Column(BigInteger, primary_key=True, nullable=False)
    service_id = Column(Integer, ForeignKey(Service.id, ondelete='CASCADE'), nullable=False, doc="Service id")

    checked = Column(Boolean, nullable=False, default=False, doc="State checked (alerts created)?")
    rtime = Column(DateTime, default=datetime.utcnow, doc="Received time")

    state = Column(Enum(*state_t.states, name='service_state'), default='UNK', nullable=False, doc='Service status')
    info = Column(UnicodeText, nullable=False, doc='Service info')

    service = relationship(Service, foreign_keys=service_id, backref=backref('states', passive_deletes=True))

    __table_args__ = (
        Index('idx_serviceid_rtime', service_id, rtime),
        Index('idx_checked', checked)
    )

    @property
    def prev(self):
        # FIXME: learn how to apply .limit() to the relationship() query and uncomment the definition below
        return object_session(self)\
            .query(ServiceState)\
            .filter(
                ServiceState.id < self.id,
                ServiceState.service_id == self.service_id
            )\
            .order_by(ServiceState.id.desc())\
            .first()

    # prev = relationship("ServiceState", viewonly=True, uselist=False,
    #                                  primaryjoin=and_(
    #                                      remote(id) < foreign(id),
    #                                      remote(service_id) == foreign(service_id)
    #                                  ), order_by=id.desc(), doc="Previous state, if any")


latest_state = select([func.max(ServiceState.id)]). \
    where(Service.id == ServiceState.service_id). \
    correlate(Service). \
    as_scalar()
Service.state = relationship(ServiceState, viewonly=True, uselist=False,
                             primaryjoin=and_(
                                 ServiceState.id == latest_state,
                                 ServiceState.service_id == Service.id
                             ), doc="Current service state")


class Alert(Base):
    """ Reported alerts """
    __tablename__ = 'alerts'

    id = Column(BigInteger, primary_key=True, nullable=False)
    server_id = Column(Integer, ForeignKey(Server.id, ondelete='CASCADE'), nullable=True, doc="Server id")
    service_id = Column(Integer, ForeignKey(Service.id, ondelete='CASCADE'), nullable=True, doc="Service id")

    reported = Column(Boolean, nullable=False, default=False, doc="Alert reported (notifications sent)?")
    ctime = Column(DateTime, default=datetime.utcnow, doc="Creation time")

    channel = Column(String(32), nullable=False, doc="Alert channel")
    event = Column(String(32), nullable=False, doc="Alert event")
    message = Column(UnicodeText, nullable=False, doc="Alert details")

    server = relationship(Server, foreign_keys=server_id, backref=backref('alerts', passive_deletes=True, order_by=id.desc()))
    service = relationship(Service, foreign_keys=service_id, backref=backref('alerts', passive_deletes=True, order_by=id.desc()))

    __table_args__ = (
        Index('idx_reported', reported),
    )

    def __unicode__(self):
        server_service = u' '.join(filter(lambda x: x is not None, [
            unicode(self.server) if self.server else None,
            unicode(self.service) if self.service else None,
        ]))
        return u''.join([
            # server `service`:
            u'{}: '.format(server_service) if server_service else u'',
            # [channel/event]
            u'[{}/{}] '.format(self.channel, self.event),
            # message
            self.message
        ])



# -*- coding: utf-8 -*-

from time import sleep
import unittest
import os
from datetime import datetime

from . import ApplicationTest
from overc.lib.db import models
from overc.lib.alerts import AlertPlugin
from overc.lib.supervise import supervise_once


class ApiTest(ApplicationTest, unittest.TestCase):
    """ Test API: /api """

    def send_service_status(self, server, services, period=60):
        return self.test_client.jsonapi('POST', '/api/set/service/status', {
            'server': server,
            'period': period,
            'services': services
        })

    def send_alerts(self, server, alerts):
        return self.test_client.jsonapi('POST', '/api/set/alerts', {
            'server': server,
            'alerts': alerts,
        })

    def assertServices(self, server, expected):
        """ Helper to test for services' states """
        # Get services
        services = list(server.services)
        services.sort(lambda a, b: cmp(a.id, b.id))
        self.assertEqual(len(services), len(expected))

        # Iterate
        for i, service in enumerate(services):
            self.assertEqual(service.id, expected[i]['id'])
            self.assertEqual(service.period, expected[i]['period'])
            self.assertEqual(service.name, expected[i]['name'])
            self.assertEqual(service.title, expected[i]['title'])

            if expected[i]['state'] is None:
                self.assertIsNone(service.state)
            else:
                self.assertEqual(service.state.id, expected[i]['state']['id'])
                self.assertEqual(service.state.checked, expected[i]['state']['checked'])
                self.assertIsInstance(service.state.rtime, datetime)
                self.assertEqual(service.state.state, expected[i]['state']['state'])
                self.assertEqual(service.state.info, expected[i]['state']['info'])

    def assertAlerts(self, server, expected):
        """ Helper to test for server alerts """
        # Get alerts
        alerts = list(server.alerts)
        alerts.sort(lambda a, b: cmp(a.id, b.id))
        self.assertEqual(len(alerts), len(expected))

        # Iterate
        for i, alert in enumerate(alerts):
            self.assertEqual(alert.id, expected[i]['id'])
            self.assertEqual(alert.service_id, expected[i]['service_id'])
            self.assertEqual(alert.reported, expected[i]['reported'])
            self.assertIsInstance(alert.ctime, datetime)
            self.assertEqual(alert.channel, expected[i]['channel'])
            self.assertEqual(alert.event, expected[i]['event'])
            self.assertEqual(alert.message, expected[i]['message'])

    def test_ping(self):
        """ Test /api/ping """
        # First ping: server is created
        res, rv = self.test_client.jsonapi('POST', '/api/ping', { 'server': { 'name': 'localhost', 'key': '1234' } })
        self.assertEqual(rv.status_code, 200)

        # Server was created
        server = self.db.query(models.Server).filter(models.Server.id == 1).first()
        self.assertIsNotNone(server)
        self.assertEqual(server.name, 'localhost')
        self.assertEqual(server.title, 'localhost')
        self.assertEqual(server.key, '1234')

        # Second ping: still fine
        res, rv = self.test_client.jsonapi('POST', '/api/ping', {'server': {'name': 'localhost', 'key': '1234'}})
        self.assertEqual(rv.status_code, 200)

        # Ping with a wrong key: error
        res, rv = self.test_client.jsonapi('POST', '/api/ping', {'server': {'name': 'localhost', 'key': '1__4'}})
        self.assertEqual(rv.status_code, 403)

    def test_service_status(self):
        """ Test /api/set/service/status """

        # Test validation: 'server'
        res, rv = self.send_service_status({'name': 'localhost'}, [])
        self.assertEqual(rv.status_code, 400)
        self.assertDictEqual(res, {'error': 'Data: "server.key" should be a string'})

        # Send the first status
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            dict(name='app', state='OK', info='up 30s'),
            dict(name='cpu', state='OK', info='50% ok'),
            dict(name='que', state='HEY', info='2 ok'), # unsupported state
        ])
        self.assertEqual(rv.status_code, 200)

        # A server should've been created
        server = self.db.query(models.Server).filter(models.Server.id == 1).first()
        self.assertEqual(server.name, 'localhost')
        self.assertEqual(server.title, 'localhost')
        self.assertEqual(server.key, '1234')

        # Services should've been created
        self.assertServices(server, [
            dict(id=1, period=60, name='app', title=u'app', state=dict(id=1, checked=False, state='OK', info='up 30s')),
            dict(id=2, period=60, name='cpu', title=u'cpu', state=dict(id=2, checked=False, state='OK', info='50% ok')),
            dict(id=3, period=60, name='que', title=u'que', state=dict(id=3, checked=False, state='UNK', info='2 ok (sent unsupported state: "HEY")')),
        ])

        # Update with more status
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            dict(name='app', state='OK', info='up 40s'),
            dict(name='cpu', state='OK', info='30% ok'),
            dict(name='que', state='OK', info='3 ok'),
            dict(name='moo', state='OK', info=':)')
        ])
        self.assertEqual(rv.status_code, 200)

        # Test Services
        server = self.db.query(models.Server).filter(models.Server.id == 1).first()
        self.assertServices(server, [
            dict(id=1, period=60, name='app', title=u'app', state=dict(id=4, checked=False, state='OK', info='up 40s')),
            dict(id=2, period=60, name='cpu', title=u'cpu', state=dict(id=5, checked=False, state='OK', info='30% ok')),
            dict(id=3, period=60, name='que', title=u'que', state=dict(id=6, checked=False, state='OK', info='3 ok')),
            dict(id=4, period=60, name='moo', title=u'moo', state=dict(id=7, checked=False, state='OK', info=':)')),
        ])

        # Try to update with an invalid server key
        res, rv = self.send_service_status({'name': 'localhost', 'key': '____'}, [])
        self.assertEqual(rv.status_code, 403)
        self.assertDictEqual(res, {'error': 'Invalid server key'})

        # Finally, try to report a single service multiple times and make sure it does not make duplicates
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            dict(name='test', state='OK', info='1'),
            dict(name='test', state='OK', info='2', period=13),
        ])
        self.assertEqual(rv.status_code, 200)

        server = self.db.query(models.Server).filter(models.Server.id == 1).first()
        self.assertServices(server, [
            dict(id=1, period=60, name='app', title=u'app', state=dict(id=4, checked=False, state='OK', info='up 40s')),
            dict(id=2, period=60, name='cpu', title=u'cpu', state=dict(id=5, checked=False, state='OK', info='30% ok')),
            dict(id=3, period=60, name='que', title=u'que', state=dict(id=6, checked=False, state='OK', info='3 ok')),
            dict(id=4, period=60, name='moo', title=u'moo', state=dict(id=7, checked=False, state='OK', info=':)')),
            dict(id=5, period=13, name='test', title=u'test', state=dict(id=9, checked=False, state='OK', info='2')),
        ])

    def test_alerts(self):
        """ Test /api/set/alerts """
        # Send the first bunch of alerts
        res, rv = self.send_alerts({ 'name': 'localhost', 'key': '1234' }, [
            dict(message='Server lags'),
            dict(message='Too much logs'),
            dict(message='Service down', service='test'),
            dict(message='Service down again', service='test'),
        ])
        self.assertEqual(rv.status_code, 200)

        # Check services
        server = self.db.query(models.Server).filter(models.Server.id == 1).first()
        self.assertServices(server, [
            dict(id=1, period=None, name='test', title=u'test', state=None),
        ])

        # Check alerts
        self.assertAlerts(server, [
            dict(id=1, service_id=None, reported=False, channel='api', event='alert', message=u'Server lags'),
            dict(id=2, service_id=None, reported=False, channel='api', event='alert', message=u'Too much logs'),
            dict(id=3, service_id=1, reported=False, channel='api', event='alert', message=u'Service down'),
            dict(id=4, service_id=1, reported=False, channel='api', event='alert', message=u'Service down again'),
        ])

    def test_supervisor(self): # TODO: use freezegun
        """ Test how alerts are created when the service state changes """
        # Prerequisites
        def overc_readlog():
            """ Read overc.log and remove """
            overc_log = '/tmp/overc.log'
            if not os.path.exists(overc_log):
                return None
            try:
                with open(overc_log, 'r') as f:
                    return f.read().encode('utf-8')
            finally:
                os.unlink(overc_log)
        overc_readlog()

        # Dry run first
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))
        self.assertIsNone(overc_readlog())

        # Test a: OK
        # Nothing should be alerted or reported
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'OK', 'info': 'hey1'},
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))
        self.assertIsNone(overc_readlog())

        # Test b: UNK
        # UNKs should always be reported
        # Also, that's the first reported state, and it should only be reported once
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'b', 'state': 'BULLSHIT', 'info': 'hey2'},
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (1, 1)) # 1 alert
        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost b: '
            u'[service:state/changed] '
            u'State changed: "(?)" -> "UNK"'
            '\n'
            u'Current: UNK: hey2 (sent unsupported state: "BULLSHIT")'
            '\n'
        )

        # Test a: OK -> WARN
        # State change should be reported
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'WARN', 'info': 'hey3'},
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (1, 1)) # 1 alert
        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost a: '
            u'[service:state/changed] '
            u'State changed: "OK" -> "WARN"'
            '\n'
            u'Current: WARN: hey3'
            '\n'
        )

        # Test a: WARN -> WARN
        # State not changed, no alert
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'WARN', 'info': 'hey4'},
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))  # no alerts

        # Test a: WARN -> OK
        # State change should be reported
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'OK', 'info': 'hey5'},
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (1, 1))  # 1 alert
        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost a: '
            u'[service:state/changed] '
            u'State changed: "WARN" -> "OK"'
            '\n'
            u'Current: OK: hey5'
            '\n'
        )

        # Test a: OK -> (timeout)
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'OK', 'info': 'hey6'},
        ], period=1)
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))  # all ok

        sleep(2)

        self.assertEqual(supervise_once(self.app, self.db), (1, 1))  # 1 alert
        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost a: '
            u'[plugin/offline] '
            u'Monitoring plugin offline'
            '\n'
            u'Current: OK: hey6'
            '\n'
        )

        # Still offline, it should not report again
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))  # no alerts

        # Test a: OK back again
        res, rv = self.send_service_status({'name': 'localhost', 'key': '1234'}, [
            {'name': 'a', 'state': 'OK', 'info': 'hey7'},
        ], period=60)
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (1, 1))  # 1 alert
        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost a: '
            u'[plugin/online] '
            u'Monitoring plugin back online'
            '\n' +
            u'Current: OK: hey7'
            '\n'
        )

        # Still online, it should not report again
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))  # no alerts



        # Test manual alerts
        self.send_alerts({'name': 'localhost', 'key': '1234'}, [
            dict(message='Server lags'),
            dict(message='Service down again', service='test'),
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 2))  # 2 alerts

        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost: '
            u'[api/alert] '
            u'Server lags'
            '\n'
            u'localhost test: '
            u'[api/alert] '
            u'Service down again'
            '\n'
        )

        # Should not report again
        self.assertEqual(supervise_once(self.app, self.db), (0, 0))  # no alerts



        # See how a failing script behaves
        self.app.app.config['ALERT_PLUGINS'].append(
            AlertPlugin(name='error', cwd=self.app.app.instance_path, command='./alert.d/error.sh')
        )

        self.send_alerts({'name': 'localhost', 'key': '1234'}, [
            dict(message='Server lags'),
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 1))  # 1 alert, but 1 additional fatal

        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost: '
            u'[api/alert] '
            u'Server lags'
            '\n'
            u'Alert plugin `error` failed: Command \'./alert.d/error.sh\' returned non-zero exit status 1'
            '\n'
        )

        # See how a missing script behaves
        self.app.app.config['ALERT_PLUGINS'].pop()
        self.app.app.config['ALERT_PLUGINS'].append(
            AlertPlugin(name='nonexistent', cwd=self.app.app.instance_path, command='./alert.d/nonexistent')
        )

        self.send_alerts({'name': 'localhost', 'key': '1234'}, [
            dict(message='Server lags'),
        ])
        self.assertEqual(res['ok'], 1)
        self.assertEqual(supervise_once(self.app, self.db), (0, 1))  # 1 alert, but 1 additional fatal

        self.assertMultiLineEqual(
            overc_readlog(),
            u'localhost: '
            u'[api/alert] '
            u'Server lags'
            '\n'
            u'Alert plugin `nonexistent` failed: [Errno 2] No such file or directory'
            '\n'
        )

    def test_unicode(self):
        """ Check how API handles unicode """
        res, rv = self.send_service_status({'name': u'сервер', 'key': u'ключ'}, [
            { 'name': u'сервис', 'state': u'ХОРОШО', 'info': u'Всё отлично' }
        ])
        self.assertEqual(rv.status_code, 200)



#! /usr/bin/env python

import pkg_resources

__author__ = "Mark Vartanyan"
__email__ = "kolypto@gmail.com"
__version__ = pkg_resources.get_distribution('overc').version

from overcli.commands import main

if __name__ == '__main__':
    main()



import os

from flask.testing import FlaskClient
from flask import request, json

from overc import OvercApplication
from overc.lib.db import models



class CustomFlaskClient(FlaskClient):
    """ Custom Flask client for tests """

    def jsonapi(self, method, uri, body=None, **kwargs):
        """ Call a JSON API method
        :param method: HTTP method name
        :type method: str
        :param uri: URI
        :type uri: str
        :param body: Data to submit
        :type body: *
        :return: The response tuple
        :rtype: (basestring|dict, flask.wrappers.Response)
        """
        # Send
        rv = self.open(
            uri,
            method=method,
            content_type='application/json',
            data=json.dumps(body or {}),
            **kwargs
        )

        # JSON?
        if rv.mimetype == 'application/json':
            return json.loads(rv.get_data()), rv

        # Non-JSON
        return rv.get_data(), rv


class ApplicationTest(object):
    """ Mixin to initialize the app """

    def setUp(self):
        # Load config
        app_config_file = 'tests/data/overc-server/server.ini'
        app_config = OvercApplication.loadConfigFile(app_config_file)

        # Init app
        self.instance_path = app_config['INSTANCE_PATH']
        self.app = OvercApplication(__name__, self.instance_path, app_config)
        self.app.app.config.update(TESTING=True)

        # Customize
        self.app.app.test_client_class = CustomFlaskClient

        # Reset DB
        models.Base.metadata.drop_all(bind=self.app.db_engine)
        models.Base.metadata.create_all(bind=self.app.db_engine)

    def tearDown(self):
        # Close & reset DB
        self.app.db.close()
        models.Base.metadata.drop_all(bind=self.app.db_engine)

    #region Helpers

    @property
    def db(self):
        """ Get DB session
            :rtype: sqlalchemy.orm.session.Session
        """
        return self.app.db

    @property
    def test_client(self):
        """ Get test client
        :rtype: CustomFlaskClient
        """
        return self.app.app.test_client()

    #endregion



import os
from ConfigParser import ConfigParser

from flask import Flask, Request
from flask.ctx import _AppCtxGlobals

from overc import __version__
from overc.src.init import init_db_engine, init_db_session_for_flask
from overc.lib.alerts import AlertPlugin

class OvercFlask(Flask):
    """ Custom Flask """

class OvercApplication(object):
    """ OverC Application """
    version = __version__

    @staticmethod
    def loadConfigFile(filename):
        """ Load OverC config file and parse it
        Any config option <name> can be overriden from environment: `OVERC_<name>`

        :param filename: Config file path
        :type filename: str
        :return: Configuration
        :rtype: dict
        """
        # Default config
        app_config = dict(
            DEBUG=False,
            TESTING=False,
            VERBOSE=0,
            SERVER_NAME=None,
            APPLICATION_ROOT=None,
            PREFERRED_URL_SCHEME='http',

            INSTANCE_PATH='/',
            DATABASE='mysql://user:pass@127.0.0.1/overc',
            ALERT_PLUGINS=[],
        )

        # Load config
        if not os.path.exists(filename):
            raise OSError('Config file does not exist: {}'.format(filename))
        app_config['INSTANCE_PATH'] = os.path.dirname(os.path.realpath(filename))

        ini = ConfigParser()
        ini.read(filename)

        # Parse: [overc]
        if ini.has_section('overc'):
            app_config['DATABASE'] = ini.get('overc', 'database')
            app_config['LOGLEVEL'] = ini.get('overc', 'loglevel')

        # Parse: [alert:*]
        for s in ini.sections():
            if s.startswith('alert:'):
                app_config['ALERT_PLUGINS'].append(
                    AlertPlugin(
                        name=s.split(':', 1)[1],
                        cwd=app_config['INSTANCE_PATH'],
                        command=ini.get(s, 'command')
                    )
                )

        # Override from environment
        for name, value in os.environ.items():
            if name.startswith('OVERC_'):
                app_config[name[len('OVERC_'):]] = value

        # Finish
        return app_config

    def __init__(self, import_name, instance_path, config):
        """ Initialize app
        :param import_name: Declaring module nmae
        :type import_name: str
        :param instance_path: Application instance path
        :type instance_path: str
        :param config: Application configuration object
        :type config: dict
        """
        # Init app
        self.app = OvercFlask(import_name,
            template_folder='templates',
            instance_path=instance_path,
            static_folder='static',
            static_url_path='/static'
        )
        self.app.config.update(config)
        self.app.debug = config.get('DEBUG', False)

        # Init DB
        self.db_engine = init_db_engine(self.app.config['DATABASE'])
        self.db = init_db_session_for_flask(self.db_engine, self.app)

        # Globals
        class DignioAppCtxGlobals(_AppCtxGlobals):
            """ Flask `g` overrides """
            app = self
            db = self.db
        self.app.app_ctx_globals_class = DignioAppCtxGlobals

        # Blueprints
        from .bps import api
        from .bps import ui
        self.app.register_blueprint(api.bp, url_prefix='/api')
        self.app.register_blueprint(ui.bp, url_prefix='/ui')



import json, urlparse, urllib2, base64

class Overclient(object):
    """ OverC API client """

    def __init__(self, url, server_name, server_key):
        """ Initialize the client
        :param url: OverC server URL
        :type url: str
        :param server_name: Server identification: name
        :type server_name: str
        :param server_key: Server identification: key
        :type server_key: str
        """
        # Parse URL
        headers = {}
        p = urlparse.urlsplit(url, 'http')
        if p.username and p.password:
            # Prepare header
            headers['Authorization'] = b'Basic ' + base64.b64encode(p.username + b':' + p.password)
            # Remove authentication info since urllib2.Request() does not understand it
            url = urlparse.urlunsplit((p.scheme, p.netloc.split('@',1)[1], p.path, p.query, p.fragment))

        # Put
        self._url = url
        self._headers = headers
        self._server_id = {'name': server_name, 'key': server_key}

    def _jsonpost(self, path, data=None):
        """ Execute an API method
        :param path: Method URI
        :type path: str
        :param data: Data to POST (optional)
        :type data: dict
        :return: Response
        :rtype: dict
        :exception urllib2.URLError: Connection errors
        """
        # Prepare
        url = urlparse.urljoin(self._url, path)
        req = urllib2.Request(url)
        if data:
            req.add_header('Content-Type', 'application/json')
        for name, value in self._headers.items():
            req.add_header(name, value)

        # Request
        response = urllib2.urlopen(req, json.dumps(data) if data else None)

        # Read
        res_str = response.read()
        res = json.loads(res_str)
        return res

    def ping(self):
        """ Test connection
        :exception urllib2.URLError: Connection errors
        """
        return self._jsonpost('/api/ping', {
            'server': self._server_id
        })

    def set_service_status(self, period, services):
        """ Send services' status
        :param period: Promised reporting period
        :type period: int
        :param services: List of services to report
        :type services: list
        :exception urllib2.URLError: Connection errors
        """
        return self._jsonpost('/api/set/service/status', {
            'server': self._server_id,
            'period': period,
            'services': services
        })

    def set_alerts(self, alerts):
        """ Send alerts
        :param alerts: Alerts to report
        :type alerts: list
        :exception urllib2.URLError: Connection errors
        """
        return self._jsonpost('/api/set/alerts', {
            'server': self._server_id,
            'alerts': alerts
        })



from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker

def init_db_engine(connect):
    """ Init Sqlalchemy engine
    :param connect: Database connection string
    :type connect: str
    :rtype: sqlalchemy.engine.Engine
    """
    return create_engine(connect, convert_unicode=True)

def init_db_session(engine):
    """ Init DB session
    :param engine: Engine
    :type engine: sqlalchemy.engine.Engine
    :rtype: sqlalchemy.orm.session.scoped_session
    """
    Session = scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=engine))

    # Models
    from overc.lib.db.models import Base
    Base.query = Session.query_property()
    Base.metadata.create_all(bind=engine)  # TODO: remove automatic table creation

    return Session


def init_db_session_for_flask(engine, app):
    """ Init DB session for Flask app
    :param engine: Engine
    :type engine: sqlalchemy.engine.Engine
    :param app: Flask application
    :type app: Flask
    :rtype: sqlalchemy.orm.session.scoped_session
    """
    Session = init_db_session(engine)

    # Tear-down
    @app.teardown_appcontext
    def shutdown_session(exception=None):
        Session.remove()

    return Session



import os
import time
import logging
import argparse
from ConfigParser import ConfigParser

from . import __author__, __email__, __version__
from .overclient import Overclient
from .monitor import ServicesMonitor, Service

logger = logging.getLogger(__name__)


def cmd_ping(args, overc):
    """ Ping server (api: /api/ping) """
    overc.ping()


def cmd_service_status(args, overc):
    """ Report single service status (api: /api/set/service/status) """
    overc.set_service_status(args.period, [
        { 'name': args.name, 'state': args.state, 'info': args.info }
    ])


def cmd_alert(args, overc):
    """ Report single alert (api: /api/set/alert) """
    alert = {'message': args.message}
    if args.service:
        alert['service'] = args.service
    overc.set_alerts([ alert ])


def cmd_monitor(args, overc):
    config_file = os.path.realpath(args.config)

    # Config file exists?
    cwd = os.path.dirname(config_file)
    if not os.path.exists(config_file):
        raise OSError('Config file does not exist: {}'.format(config_file))

    # Read config file
    services = []
    ini = ConfigParser()
    ini.read(config_file)

    # Overclient
    if not overc:
        overc = Overclient(
            ini.get('overc', 'server'),
            ini.get('overc', 'my-name'),
            ini.get('overc', 'my-key')
        )

    # Parse config file
    for section in ini.sections():
        # Parse [A:B] sections only
        type, name = (section.split(':', 1) + [None, None])[:2]
        if name is None:
            continue

        # Service definitions
        if section.startswith('service:'):
            services.append(Service(
                ini.getint(section, 'period'),
                name,
                cwd,
                ini.get(section, 'command')
            ))
        else:
            raise ValueError('Unknown section type: "{}"'.format(type))

    # Continuous Monitoring
    try:
        _cmd_monitor_daemon(services, overc)
    except Exception as e:
        overc.set_alerts([
            u'Monitoring failure: {}: {}'.format(e.__class__.__name__, e.message)
        ])
        raise

def _cmd_monitor_daemon(services, overc):
    """ Perform continuous monitoring """
    # First, ping the server
    overc.ping()  # exception when fails

    # Init monitor
    monitor = ServicesMonitor(services)

    while True:
        # Determine sleep time
        delay = monitor.sleep_time()
        time.sleep(delay)

        # Check
        period, service_states = monitor.check()

        # Report
        try:
            overc.set_service_status(period, service_states)
        except Exception as e:
            logger.exception('Failed to report service status!')
            # proceed

def main():
    # Arguments
    versionstr = 'v{} (c) {} <{}>'.format(__version__, __author__, __email__)
    parser = argparse.ArgumentParser(
        prog='overcli',
        description='OverC client',
        epilog=versionstr
    )
    parser.add_argument('--verbose', '-v', action='count', default=0, help='Be more verbose. -vv includes debug output')
    parser.add_argument('-s', '--server', dest='server_url', help='OverC Server URL: "http://<host>:<port>/"')
    parser.add_argument('-i', '--server-id', dest='server_id', help='Server identification: "<name>:<key>"')

    # Subcommands
    sub = parser.add_subparsers(dest='command_name', title='Command')

    cmd = sub.add_parser('ping', help='Ping server')
    cmd.set_defaults(func=cmd_ping)

    cmd = sub.add_parser('service-status', help=cmd_service_status.__doc__)
    cmd.add_argument('period', help='Reporting period')
    cmd.add_argument('name', help='Service name')
    cmd.add_argument('state', help='Service state')
    cmd.add_argument('info', nargs='?', default='', help='Additional information')
    cmd.set_defaults(func=cmd_service_status)

    cmd = sub.add_parser('alert', help=cmd_alert.__doc__)
    cmd.add_argument('--service', nargs='?', default=None, help='Service to report the alert for')
    cmd.add_argument('message', help='Alert message')
    cmd.set_defaults(func=cmd_alert)

    cmd = sub.add_parser('monitor', help=cmd_monitor.__doc__)
    cmd.add_argument('config', help='Monitoring configuration file')
    cmd.set_defaults(func=cmd_monitor)

    # Parse
    args = parser.parse_args()

    # Configure logging
    logging.basicConfig(level=[logging.WARN, logging.INFO, logging.DEBUG, logging.NOTSET][args.verbose])

    # Overclient
    if args.command_name == 'monitor':
       overc = None  # Created from the config file
    elif not args.server_url or args.server_id:
        raise RuntimeError('--server and --server-id should be specified')
    else:
        args.server_id = args.server_id.split(':', 1)
        overc = Overclient(args.server_url, *args.server_id)

    # Command
    args.func(args, overc)



import logging
import subprocess, shlex
import threading
from datetime import datetime

logger = logging.getLogger(__name__)


class Service(object):
    def __init__(self, period, name, cwd, command):
        """ Define a service to be monitored
        :param period: Test period, seconds
        :type period: int
        :param name: Service name
        :type name: str
        :param cwd: Current working directory
        :type cwd: str
        :param command: Full command path that test service state
        :type command: str
        """
        self.period = period
        self.name = name
        self.cwd = cwd
        self.command = command

        #: Plugin execution time
        self.lag = 0

        #: Timestamp when this service was tested last time
        self.last_tested = None

    def __str__(self):
        return self.name

    PERIOD_MARGIN_FACTOR = 0.8
    LAG_MARGIN_FACTOR = 3.0

    @property
    def real_period(self):
        """ Real update period, including lags and safety reserves """
        return max(self.period * self.PERIOD_MARGIN_FACTOR - self.lag * self.LAG_MARGIN_FACTOR, 0.0)

    def next_update_in(self, now):
        """ Get the relative time for the next update
        :param now: Current datetime
        :type now: datetime
        :return: Delay, seconds
        :rtype: float
        """
        # Never updated: NOW!
        if self.last_tested is None:
            return 0.0

        # Was updated
        seconds_ago = (now - self.last_tested).total_seconds()
        delay = self.real_period - seconds_ago
        return max(delay, 0.0)  # don't allow it to be negative


    def get_state(self):
        """ Execute plugin and get service's state
        :return: Process state for the API
        :rtype: dict
        :exception OSError: Failed to execute plugin
        """
        # Execute command
        try:
            process = subprocess.Popen(
                shlex.split(self.command),
                cwd=self.cwd,
                stdin=None,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT
            )
            process.wait()
        except OSError, e:
            error_msg = u'Failed to execute plugin `{}` by command `{}`: {}'.format(self.name, self.command, e.message)
            logger.exception(error_msg)
            return {
                'name': self.name,
                'state': 'UNK',
                'info': error_msg,
                'period': self.period
            }

        # Analyze the result
        try:
            # Info
            info = process.stdout.read()

            # Determine state
            try:
                state = ['OK', 'WARN', 'FAIL', 'UNK'][process.returncode]
            except IndexError:
                logger.error(u'Plugin `{}` failed with code {}: {}'.format(self.name, process.returncode, info))
                state = 'UNK'

            # Finish
            return {
                'name': self.name,
                'state': state,
                'info': unicode(info).rstrip()
            }
        finally:
            process.stdout.close()


class ServicesMonitor(object):
    def __init__(self, services):
        """ Monitor for Services
        :param services: List of Services to be monitored
        :type services: list
        """
        self.services = services

    def _check_services(self, services):
        """ Check services provided as an argument
        :param services: List of services to test
        :type services: list[Service]
        :return: `services` argument value for reporting
        :rtype: list
        """
        now = datetime.utcnow()

        # Worker
        service_states = []
        def task(service):
            # Get state, measure lag
            start = datetime.utcnow()
            state = service.get_state()
            finish = datetime.utcnow()

            # Update lag
            service.lag = (finish - start).total_seconds()

            # Add state
            service_states.append(state)
            logger.debug(u'Checked service {} (lag={}, real_period={}): last checked {} ago, state={}: {}'.format(
                service.name,
                service.lag,
                service.real_period,
                now - service.last_tested if service.last_tested else '(never)',
                state['state'], state['info']
            ))

            # Update timestamp
            service.last_tested = now

        # Run
        threads = [threading.Thread(target=task, args=(service,)) for service in services]
        for t in threads: t.start()
        for t in threads: t.join()

        return service_states

    def sleep_time(self):
        """ Determine how many seconds is it ok to sleep before any service state should be reported
        :rtype: float
        """
        now = datetime.utcnow()
        return min(service.next_update_in(now) for service in self.services)

    def check(self):
        """ Check services whose time has come, once.
        :return: (period, service_states) to be reported to the API
        :rtype: (int, dict)
        """
        # Determine which services to test
        # TODO: use a smarter algorithm to detect which services to check
        max_lag = max(service.lag for service in self.services)
        now = datetime.utcnow()
        services = [ service
                     for service in self.services
                     if service.next_update_in(now) <= max_lag
        ]
        if not services:
            return 0, []

        period = max(service.period for service in services)

        # Test them
        service_states = self._check_services(services)

        # Report
        return int(period), service_states



import os
import shlex
import subprocess
import logging

logger = logging.getLogger(__name__)


class AlertPlugin(object):
    """ Alert plugin """

    def __init__(self, name, cwd, command):
        self.name = name
        self.cwd = cwd
        self.command_str = command
        self.command = shlex.split(self.command_str)

    def send(self, message):
        """ Send a message
        :param message: Message
        :type message: unicode
        :exception OSError: plugin not found
        :exception subprocess.CalledProcessError: plugin execution error (non-zero return code)
        """
        # Execute
        process = subprocess.Popen(
            self.command,
            cwd=self.cwd,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT
        )

        try:
            # Run
            process.stdin.write(message)
            process.stdin.close()

            # Wait, analyze retcode
            process.wait()

            if process.returncode != 0:
                raise subprocess.CalledProcessError(process.returncode, self.command_str, '')
        finally:
            process.stdout.close()


def send_alert_with_plugins(alert_plugins, message):
    """ Fetch notifications config and send messages
    :param alert_plugins: List of alert plugins to use
    :type alert_plugins: list[AlertPlugin]
    :param message: Alert message to send
    :type message: unicode
    """
    # TODO: execute alert scripts in parallel!

    # Send
    plugin_failures = []
    for plugin in alert_plugins:
        try:
            plugin.send(message)
        except Exception as e:
            logger.exception('Alert plugin `{}` command failed: {}'.format(plugin.name, plugin.command_str))
            plugin_failures.append(
                'Alert plugin `{}` failed: {}'.format(plugin.name, e)
            )

    # No failures? Good!
    if not plugin_failures:
        return

    # Report plugin errors with all available plugins
    plugin_failure_message = "\n".join(plugin_failures)
    n_sent = 0
    for plugin in alert_plugins:
        try:
            plugin.send(plugin_failure_message)
            n_sent += 1
        except Exception:
            logger.exception('Failed to send fatal plugin notification!')

    # Fatal error
    if n_sent == 0:
        logger.fatal('NONE of the plugins could send the message:\n' + plugin_failure_message)



from datetime import datetime, timedelta
from logging import getLogger
from collections import defaultdict

from sqlalchemy.sql import func
from flask import Blueprint
from flask.templating import render_template
from flask.globals import g, request

from overc.lib.db import models
from overc.lib.flask.json import jsonapi

bp = Blueprint('ui', __name__, url_prefix='/ui', template_folder='templates',
               static_folder='static', static_url_path='/static'
               )
logger = getLogger(__name__)


#region UI

@bp.route('/', methods=['GET'])
def index():
    """ Index page """
    return render_template('pages/index.htm')

#endregion

#region API

@bp.route('/api/status/server')
@bp.route('/api/status/server/<server_id>')
@jsonapi
def api_status_server(server_id=None):
    """ Get all available information """
    ssn = g.db

    # Filter servers
    servers = ssn.query(models.Server)
    if server_id:
        servers = servers.filter(models.Server.id == server_id)
    servers = servers.all()

    # Count alerts for 24h
    alert_counts = ssn.query(
        models.Alert.server_id,
        models.Alert.service_id,
        func.count(models.Alert)
    ) \
        .filter(
            models.Alert.ctime >= (datetime.utcnow() - timedelta(hours=24)),
            models.Alert.server_id == server_id if server_id else True
        ) \
        .group_by(models.Alert.server_id, models.Alert.service_id) \
        .all()
    server_alerts = defaultdict(lambda: 0)
    service_alerts = defaultdict(lambda: 0)
    total_alerts = 0
    for (server_id, service_id, n) in alert_counts:
        total_alerts += n
        if service_id:
            service_alerts[service_id] += n
        elif server_id:
            server_alerts[server_id] += n

    # Test whether there are any service states not checked, which probably means the supervisor thread is no running
    last_checked = ssn.query(func.min(models.ServiceState.rtime)) \
        .filter(
            models.ServiceState.checked == False,
        ) \
        .scalar()
    supervisor_lag = (datetime.utcnow() - last_checked).total_seconds() if last_checked else 0.0

    #     Format
    return {
        'n_alerts': total_alerts,  # alerts today (for all selected servers)
        'supervisor_lag': supervisor_lag,  # Seconds ago the supervisor process last checked something
        'servers': sorted([
            {
                'id': server.id,
                'name': server.name,
                'title': server.title,
                'ip': server.ip,
                'n_alerts': server_alerts[server.id],  # alerts today, for this server
                'services': sorted([
                    {
                        'id': service.id,
                        'period': service.period,
                        'name': service.name,
                        'title': service.title,
                        'n_alerts': service_alerts[service.id],  # alerts today, for this service
                        'state': {
                            'rtime': service.state.rtime.isoformat(sep=' '),
                            'timed_out': service.timed_out,
                            'seen_ago': str(datetime.utcnow() - service.state.rtime).split('.')[0],
                            'state': service.state.state,
                            'info': service.state.info,
                        } if service.state else None
                    } for service in server.services
                ], cmp, lambda s: s['name'])
            } for server in servers
        ], cmp, lambda s: s['name'])
    }


@bp.route('/api/status/alerts/server')
@bp.route('/api/status/alerts/server/<server_id>')
@bp.route('/api/status/alerts/service/<service_id>')
@jsonapi
def api_status_alerts(server_id=None, service_id=None):
    """ Alerts for 24h """
    ssn = g.db

    dtime = timedelta(hours=int(request.args.get('hours', default=24)))

    # Load alerts
    alerts = ssn.query(models.Alert) \
        .filter(
            models.Alert.ctime >= (datetime.utcnow() - dtime),
            models.Alert.server_id == server_id if server_id else True,
            models.Alert.service_id == service_id if service_id else True
        ) \
        .order_by(models.Alert.id.desc()) \
        .all()

    # Format
    return {
        'alerts': [
            {
                'id': alert.id,
                'server': unicode(alert.server) if alert.server else None,
                'server_id': alert.server_id,
                'service': unicode(alert.service) if alert.service else None,
                'service_id': alert.service_id,

                'ctime': alert.ctime.isoformat(sep=' '),
                'channel': alert.channel,
                'event': alert.event,
                'message': alert.message
            }
            for alert in alerts
        ]
    }


#region Items

@bp.route('/api/item/server/<server_id>', methods=['DELETE'])
@jsonapi
def api_server_delete(server_id):
    """ Server CRUD: Delete """
    ssn = g.db

    server = ssn.query(models.Server).get(server_id)
    ssn.delete(server)
    ssn.commit()

    return {'ok': 1}


@bp.route('/api/item/service/<service_id>', methods=['DELETE'])
@jsonapi
def api_service_delete(service_id):
    """ Service CRUD: Delete """
    ssn = g.db

    service = ssn.query(models.Service).get(service_id)
    ssn.delete(service)
    ssn.commit()

    return {'ok': 1}

#endregion

# TODO: display app version in the UI
# TODO: API to rename servers, services
# TODO: API to test alert plugins

#endregion



import logging
from time import sleep
import os, tempfile

from overc.src.init import init_db_engine, init_db_session
from overc.lib.db import models
from overc.lib import alerts

logger = logging.getLogger(__name__)

# TODO: these routines should obtain an exclusive lock so multiple supervise processes does not issue alerts multiple times


def _check_service_states(ssn):
    """ Test all service states, raise alerts if necessary
    :param ssn: Database session
    :type ssn: sqlalchemy.orm.session.Session
    :returns: The number of new alerts reported
    :rtype: int
    """
    # Fetch all states that are not yet checked
    service_states = ssn.query(models.ServiceState)\
        .filter(models.ServiceState.checked == False)\
        .order_by(models.ServiceState.id.asc())\
        .all()

    # Check them one by one
    new_alerts = 0
    for s in service_states:
        logger.debug(u'Checking service {server}:`{service}` state #{id}: {state}'.format(id=s.id, server=s.service.server, service=s.service, state=s.state))

        # Report state changes and abnormal states
        if s.state != (s.prev.state if s.prev else 'OK'):
            ssn.add(models.Alert(
                server=s.service.server,
                service=s.service,
                channel='service:state',
                event='changed',
                message=u'State changed: "{}" -> "{}"'.format(s.prev.state if s.prev else '(?)', s.state)
            ))
            new_alerts += 1

            # In addition, report "UNK" states!
            if s.state == 'UNK' and s.prev is not None:
                ssn.add(models.Alert(
                    server=s.service.server,
                    service=s.service,
                    channel='service:state',
                    event='unk',
                    message=u'Service state unknown!'
                ))
                new_alerts += 1

        # Save
        s.checked = True
        ssn.add(s)

    # Finish
    ssn.commit()
    return new_alerts


def _check_service_timeouts(ssn):
    """ Test all services for timeouts
    :param ssn: Database session
    :type ssn: sqlalchemy.orm.session.Session
    :returns: The number of new alerts reported
    :rtype: int
    """
    # Fetch all services which have enough data
    services = ssn.query(models.Service)\
        .filter(
            models.Service.period is not None,
            models.Service.state is not None
        ).all()

    # Detect timeouts
    new_alerts = 0
    for s in services:
        # Update state
        was_timed_out = s.timed_out
        seen_ago = s.update_timed_out()

        logger.debug(u'Checking service {service}: seen_ago={seen_ago}: {timed_out}'.format(service=s, seen_ago=seen_ago, timed_out='TIMED OUT' if s.timed_out else 'ok'))

        # Changed?
        if was_timed_out != s.timed_out:
            alert = models.Alert(
                server=s.server,
                service=s,
                channel='plugin'
            )
            if s.timed_out:
                alert.event = 'offline'
                alert.message = u'Monitoring plugin offline'
            else:
                alert.event = 'online'
                alert.message = u'Monitoring plugin back online'

            ssn.add(alert)
            ssn.add(s)
            new_alerts += 1

    # Finish
    ssn.commit()
    return new_alerts


def _send_pending_alerts(ssn, alert_plugins):
    """ Send pending alerts
    :param ssn: Database session
    :type ssn: sqlalchemy.orm.session.Session
    :param alert_plugins: Application config for alerts
    :type alert_plugins: list[alerts.AlertPlugin]
    :returns: The number of alerts sent
    :rtype: int
    """
    # Fetch all alerts which were not reported
    pending_alerts = ssn.query(models.Alert)\
        .filter(models.Alert.reported == False)\
        .all()

    # Report them one by one
    for a in pending_alerts:
        logger.debug(u'Sending alert #{id}: server={server}, service={service}, [{channel}/{event}]'.format(id=a.id, server=a.server, service=a.service, channel=a.channel, event=a.event))

        # Prepare alert message
        alert_message = unicode(a) + "\n"
        if a.service and a.service.state:
            s = a.service.state
            alert_message += u"Current: {}: {}\n".format(s.state, s.info)

        # Potential exceptions are handled & logged down there
        alerts.send_alert_with_plugins(alert_plugins, alert_message)
        a.reported = True
        ssn.add(a)

    # Finish
    ssn.commit()
    return len(pending_alerts)


def supervise_once(app, ssn):
    """ Perform all background actions once:

    * Check service states
    * Check for service timeouts
    * Send alerts

    :param app: Application
    :type app: OvercApplication
    :returns: (New alerts created, Alerts sent)
    :rtype: (int, int)
    """
    # Act
    new_alerts, sent_alerts = 0, 0
    new_alerts += _check_service_states(ssn)
    new_alerts += _check_service_timeouts(ssn)
    sent_alerts = _send_pending_alerts(ssn, app.app.config['ALERT_PLUGINS'])

    # Finish
    logger.debug('Supervise loop finished: {} new alerts, {} sent alerts'.format(new_alerts, sent_alerts))
    return new_alerts, sent_alerts





import signal, errno
from contextlib import contextmanager
import fcntl

@contextmanager
def flock_timeout(filename, seconds): # TODO: if this works, move to lib/
    original_handler = signal.signal(signal.SIGALRM, lambda signum, frame: None)
    try:
        signal.alarm(seconds)
        with open(filename, 'w') as f:
            try:
                fcntl.flock(f.fileno(), fcntl.LOCK_EX)
                yield
            except IOError as e:
                if e.errno != errno.EINTR:
                    raise e  # Some problem
                else:
                    yield # TODO: Lock timed out. Throw a special exception!
            finally:
                fcntl.flock(f.fileno(), fcntl.LOCK_UN)
    finally:
        signal.alarm(0)
        signal.signal(signal.SIGALRM, original_handler)


def supervise_loop(app):
    """ Supervisor main loop which performs background actions
    :param app: Application
    :type app: OvercApplication
    """
    db_engine = init_db_engine(app.app.config['DATABASE'])
    Session = init_db_session(db_engine)

    lockfile = os.path.join(tempfile.gettempdir(), 'overc.lock')

    while True:
        sleep(3)

        # Locking
        with flock_timeout(lockfile, seconds=2):
            # Supervise
            ssn = Session()
            try:
                # TODO: receive notifications from the API for immediate supervision
                supervise_once(app, ssn)
            except Exception:
                logger.exception('Supervise loop error')
            finally:
                Session.remove()



from functools import wraps
from flask import request, jsonify, make_response
from werkzeug.exceptions import HTTPException

def json_response(res, code=200):
    """ Make up a Response object from data
    :param res: Response data
    :type res: *
    :param code: Response code
    :type code: int
    :rtype: flask.Response
    """
    response = make_response(jsonify(res), code)
    return response

def jsonapi(f):
    """ Declare a view as JSON API method """
    @wraps(f)
    def wrapper(*args, **kwargs):
        # Invoke
        try:
            res = f(*args, **kwargs)
        except HTTPException as e:
            return json_response({'error': e.description}, e.code)
        except AssertionError as e:
            return json_response({ 'error': e.message }, 400)

        # Tuple response
        c = 200
        if isinstance(res, tuple):
            result, c = res

        # Finish
        return json_response(res, c)
    return wrapper



import os
from datetime import datetime
import unittest
from freezegun import freeze_time

from overcli.monitor import Service, ServicesMonitor


class MonitorTest(unittest.TestCase):
    """ Test overclient monitor """

    def assertServiceStates(self, service_states, expected_service_states):
        """ Test that the two service states are equal """
        f = lambda a, b: cmp(a['name'], b['name'])
        service_states.sort(f)
        expected_service_states.sort(f)
        self.assertListEqual(service_states, expected_service_states)

    def test_monitor(self):
        cwd = os.path.realpath('tests/data/overcli-config')

        # Create the monitor
        services = {
            'app': Service(15, 'app', cwd, './plugin.d/app.sh'),
            'que': Service(30, 'que', cwd, './plugin.d/que.sh'),
            'cpu': Service(30, 'cpu', cwd, './plugin.d/cpu.sh'),
            'echo': Service(15, 'echo', cwd, 'echo 1'),
        }
        monitor = ServicesMonitor(services.values())

        # Make sure it adjusted real periods
        self.assertAlmostEqual(services['app'].real_period, 15*0.8, delta=0.5)
        self.assertAlmostEqual(services['que'].real_period, 30*0.8, delta=0.5)
        self.assertAlmostEqual(services['cpu'].real_period, 30*0.8, delta=0.5)

        # Test sleep time
        self.assertEqual(monitor.sleep_time(), 0.0)  # should test immediately

        with freeze_time('2014-01-01 00:00:00'):
            # Now, test all services
            period, service_states = monitor.check()

            # Should have tested all of them
            self.assertAlmostEqual(period, 30.0, delta=0.1)  # period = max of all
            self.assertServiceStates(service_states, [
                {'name': 'app', 'state': 'OK', 'info': u'Running fine'},
                {'name': 'que', 'state': 'UNK', 'info': u'5 items'},
                {'name': 'cpu', 'state': 'WARN', 'info': u'50%'},
                {'name': 'echo', 'state': 'OK', 'info': u'1'},
            ])

            # Sleep time should be 15
            self.assertAlmostEqual(monitor.sleep_time(), 15.0*0.8, delta=0.5)

        with freeze_time('2014-01-01 00:00:11'): # 1 sec earlier
            self.assertAlmostEqual(monitor.sleep_time(), 1.0, delta=0.5)

            # No tests should be run
            period, service_states = monitor.check()
            self.assertEqual(period, 0)
            self.assertEqual(service_states, [])

        with freeze_time('2014-01-01 00:00:12'):
            self.assertEqual(monitor.sleep_time(), 0.0)

            # Run
            period, service_states = monitor.check()

            # Only `app` and `echo` should be tested
            self.assertAlmostEqual(period, 15.0, delta=0.1)
            self.assertServiceStates(service_states, [
                {'name': 'app', 'state': 'OK', 'info': u'Running fine'},
                {'name': 'echo', 'state': 'OK', 'info': u'1'},
            ])

            # Sleep time should be 15
            self.assertAlmostEqual(monitor.sleep_time(), 15.0 * 0.8, delta=0.5)

        with freeze_time('2014-01-01 00:00:24'):
            self.assertEqual(monitor.sleep_time(), 0.0)

            # Run
            period, service_states = monitor.check()

            # Should test all of them
            self.assertAlmostEqual(period, 30.0, delta=0.1)
            self.assertServiceStates(service_states, [
                {'name': 'app', 'state': 'OK', 'info': u'Running fine'},
                {'name': 'que', 'state': 'UNK', 'info': u'5 items'},
                {'name': 'cpu', 'state': 'WARN', 'info': u'50%'},
                {'name': 'echo', 'state': 'OK', 'info': u'1'},
            ])

            # Sleep time should be 15
            self.assertAlmostEqual(monitor.sleep_time(), 15.0 * 0.8, delta=0.5)

    def test_monitor_lag(self):
        """ Test monitoring of lagging tasks """
        cwd = os.path.realpath('tests/data/overcli-config')

        # Create the monitor
        services = {
            'lag1': Service(15, 'lag1', cwd, './plugin.d/lag.sh'),
            'lag2': Service(15, 'lag2', cwd, './plugin.d/lag.sh'),
            'lag3': Service(15, 'lag3', cwd, './plugin.d/lag.sh'),
            'lag4': Service(15, 'lag4', cwd, './plugin.d/lag.sh'),
        }
        monitor = ServicesMonitor(services.values())

        # Test all services, measure the time
        start = datetime.utcnow()
        period, service_states = monitor.check()
        finish = datetime.utcnow()
        finished_in = (finish - start).total_seconds()

        # Each task lags ~3sec, but they should've been run in parallel
        self.assertEqual(period, 15.0)
        self.assertEqual(len(service_states), 4)
        self.assertAlmostEqual(finished_in, 3.0, delta=1.0)

        # Lags should've been updated
        for s in services.values():
            self.assertAlmostEqual(s.lag, 3.0, delta=1.0)



# -*- coding: utf-8 -*-

import unittest
from freezegun import freeze_time

from . import ApplicationTest
from overc.lib.db import models

class UITest(ApplicationTest, unittest.TestCase):
    """ Test UI """

    def assertItemsCount(self, servers, services, alerts):
        """ Count items in the DB """
        self.assertEqual(self.db.query(models.Server).count(), servers)
        self.assertEqual(self.db.query(models.Service).count(), services)
        self.assertEqual(self.db.query(models.Alert).count(), alerts)

    @freeze_time('2014-01-01 00:00:00')
    def test_api_status_server(self):
        """ Test /api/status/server/:server_id """
        # First, report data about two servers
        res, rv = self.test_client.jsonapi('POST', '/api/set/service/status', {
            'server': {'name': 'a.example.com', 'key': '1234'},
            'period': 60,
            'services': [
                {'name': 'app', 'state': 'OK', 'info': 'Fine'},
                {'name': 'db', 'state': 'OK', 'info': 'Available'},
            ]
        })
        self.assertEqual(rv.status_code, 200)
        res, rv = self.test_client.jsonapi('POST', '/api/set/service/status', {
            'server': {'name': 'b.example.com', 'key': '9876'},
            'period': 90,
            'services': [
                {'name': 'nginx', 'state': 'OK', 'info': 'Running'},
            ]
        })
        self.assertEqual(rv.status_code, 200)

        # Send some alerts
        res, rv = self.test_client.jsonapi('POST', '/api/set/alerts', {
            'server': {'name': 'a.example.com', 'key': '1234'},
            'alerts': [
                {'message': 'Something wrong'},
                {'message': 'Something very wrong'},
                {'message': 'Service too warm', 'service': 'db'},
            ]
        })
        self.assertEqual(rv.status_code, 200)

        # Now test the API: get all services' state
        res, rv = self.test_client.jsonapi('GET', '/ui/api/status/server')
        self.assertEqual(rv.status_code, 200)
        self.assertIn('servers', res)
        self.assertEqual(len(res['servers']), 2)

        # Test server 1
        server = res['servers'][0]
        services = server.pop('services')
        self.assertDictEqual(server, {'id': 1, 'name': 'a.example.com', 'title': u'a.example.com', 'ip': None, 'n_alerts': 2})
        self.assertEqual(services[0], {'id': 1, 'period': 60, 'name': 'app',   'title': u'app',   'n_alerts': 0,
                                       'state': {'rtime': '2014-01-01 00:00:00', 'timed_out': False, 'seen_ago': '0:00:00', 'state': 'OK', 'info': 'Fine'}})
        self.assertEqual(services[1], {'id': 2, 'period': 60, 'name': 'db',    'title': u'db',    'n_alerts': 1,
                                       'state': {'rtime': '2014-01-01 00:00:00', 'timed_out': False, 'seen_ago': '0:00:00', 'state': 'OK', 'info': 'Available'}})

        # Test server 2
        server = res['servers'][1]
        services = server.pop('services')
        self.assertEqual(services[0], {'id': 3, 'period': 90, 'name': 'nginx', 'title': u'nginx', 'n_alerts': 0,
                                       'state': {'rtime': '2014-01-01 00:00:00', 'timed_out': False, 'seen_ago': '0:00:00', 'state': 'OK', 'info': 'Running'}})

        # Now try to load a single server: /:server_id
        res, rv = self.test_client.jsonapi('GET', '/ui/api/status/server/2')
        # Test server 2
        server = res['servers'][0]
        services = server.pop('services')
        self.assertEqual(services[0], {'id': 3, 'period': 90, 'name': 'nginx', 'title': u'nginx', 'n_alerts': 0,
                                       'state': {'rtime': '2014-01-01 00:00:00', 'timed_out': False, 'seen_ago': '0:00:00', 'state': 'OK', 'info': 'Running'}})

    @freeze_time('2014-01-01 00:00:00')
    def test_api_status_alerts(self):
        """ Test /api/status/alerts/ """
        # Report alerts from two servers
        res, rv = self.test_client.jsonapi('POST', '/api/set/alerts', {
            'server': {'name': 'a.example.com', 'key': '1234'},
            'alerts': [
                {'message': 'o_O',},
                {'message': 'Hurts',},
                {'message': 'Sick', 'service': 'a'},
            ]
        })
        self.assertEqual(rv.status_code, 200)

        res, rv = self.test_client.jsonapi('POST', '/api/set/alerts', {
            'server': {'name': 'b.example.com', 'key': '9876'},
            'alerts': [
                {'message': 'WOW', },
            ]
        })
        self.assertEqual(rv.status_code, 200)

        # Alerts
        alerts = [
            None,
            {'id': 1, 'server': 'a.example.com', 'server_id': 1, 'service': None, 'service_id': None, 'ctime': '2014-01-01 00:00:00', 'channel': 'api', 'event': 'alert', 'message': 'o_O'},
            {'id': 2, 'server': 'a.example.com', 'server_id': 1, 'service': None, 'service_id': None, 'ctime': '2014-01-01 00:00:00', 'channel': 'api', 'event': 'alert', 'message': 'Hurts'},
            {'id': 3, 'server': 'a.example.com', 'server_id': 1, 'service': 'a',  'service_id': 1,    'ctime': '2014-01-01 00:00:00', 'channel': 'api', 'event': 'alert', 'message': 'Sick'},
            {'id': 4, 'server': 'b.example.com', 'server_id': 2, 'service': None, 'service_id': None, 'ctime': '2014-01-01 00:00:00', 'channel': 'api', 'event': 'alert', 'message': 'WOW'}
        ]

        # Test: /ui/api/status/alerts/server
        res, rv = self.test_client.jsonapi('GET', '/ui/api/status/alerts/server')
        self.assertIn('alerts', res)
        self.assertEqual(len(res['alerts']), 4)
        self.assertDictEqual(res['alerts'][0], alerts[4])
        self.assertDictEqual(res['alerts'][1], alerts[3])
        self.assertDictEqual(res['alerts'][2], alerts[2])
        self.assertDictEqual(res['alerts'][3], alerts[1])

        # Test: /ui/api/status/alerts/server/<server_id>
        res, rv = self.test_client.jsonapi('GET', '/ui/api/status/alerts/server/1')
        self.assertIn('alerts', res)
        self.assertEqual(len(res['alerts']), 3)
        self.assertDictEqual(res['alerts'][0], alerts[3])
        self.assertDictEqual(res['alerts'][1], alerts[2])
        self.assertDictEqual(res['alerts'][2], alerts[1])

        # Test: /ui/api/status/alerts/service/<service_id>
        res, rv = self.test_client.jsonapi('GET', '/ui/api/status/alerts/service/1')
        self.assertIn('alerts', res)
        self.assertEqual(len(res['alerts']), 1)
        self.assertDictEqual(res['alerts'][0], alerts[3])


    def test_api_items_delete(self):
        """ Test items API: DELETE /api/item/* """
        # First, report something
        res, rv = self.test_client.jsonapi('POST', '/api/set/service/status', {
            'server': {'name': 'a.example.com', 'key': '1234'},
            'period': 60,
            'services': [ {'name': str(i), 'state': 'OK', 'info': 'Fine'} for i in range(1, 5) ]
        })
        self.assertEqual(rv.status_code, 200)

        res, rv = self.test_client.jsonapi('POST', '/api/set/alerts', {
            'server': {'name': 'a.example.com', 'key': '1234'},
            'alerts': [{'message': '0'}] + [ {'message': str(i), 'service': str(i)} for i in range(1, 5) ]
        })
        self.assertEqual(rv.status_code, 200)

        self.assertItemsCount(1, 4, 5)  # 1 server, 4 services, 4 service alerts, 1 server alert

        # Try to delete service
        res, rv = self.test_client.jsonapi('DELETE', '/ui/api/item/service/4')
        self.assertEqual(rv.status_code, 200)
        self.assertIsNone(self.db.query(models.Service).get(4))
        self.assertItemsCount(1, 3, 4)  # -1 service: -1 service alert

        # Try to delete server
        res, rv = self.test_client.jsonapi('DELETE', '/ui/api/item/server/1')
        self.assertEqual(rv.status_code, 200)
        self.assertIsNone(self.db.query(models.Server).get(1))
        self.assertItemsCount(0, 0, 0)  # -1 server: -3 services, -3 service alerts, -1 server alert



#!/usr/bin/env python
""" Simplistic monitoring solution that is a pleasure to use """

from setuptools import setup, find_packages

setup(
    # http://pythonhosted.org/setuptools/setuptools.html
    name='overc',
    version='1.0.5-0',
    author='Mark Vartanyan',
    author_email='kolypto@gmail.com',

    url='https://github.com/kolypto/py-overc',
    license='BSD',
    description=__doc__,
    long_description=open('README.md').read(),
    keywords=['monitoring'],

    packages=find_packages(),
    scripts=[],
    entry_points={
        'console_scripts': [
            'overcli = overcli:main',
        ]
    },

    install_requires=[
    ],
    extras_require={
        'server': [
            'flask >= 0.10.1',
            'sqlalchemy >= 0.9.6',
            'pytz >= 2014.4',
            'mysql-python >= 1.2.5'
        ],
        'tests': ['freezegun >= 0.1.18']
    },
    include_package_data=True,
    zip_safe=False,  # so we can access resources
    test_suite='tests',

    platforms='any',
    classifiers=[
        # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Flask',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        #'Programming Language :: Python :: 3',
    ],
)



from datetime import datetime
from logging import getLogger

from flask import Blueprint
from flask.globals import g, request
from werkzeug.exceptions import Forbidden

from overc.lib.db import models
from overc.lib.flask.json import jsonapi

bp = Blueprint('api', __name__, url_prefix='/api')
logger = getLogger(__name__)


def _identify_server(ssn, server_spec):
    """ Identity the server, create if missing
    :param ssn: Database session
    :type ssn: sqlalchemy.orm.session.Session
    :param server_spec: Server identification dictionary: {name: String, key: String}
    :type server_spec: dict
    :rtype: models.Server
    :exception AssertionError: Validation error
    :exception Forbidden: Invalid server key
    """

    # Input validation: server
    assert isinstance(server_spec, dict), 'Data: "server" should be a dict'
    assert 'name' in server_spec and isinstance(server_spec['name'], basestring), 'Data: "server.name" should be a string'
    assert 'key'  in server_spec and isinstance(server_spec['key'],  basestring), 'Data: "server.key" should be a string'

    # Identify server or create
    server = ssn.query(models.Server).filter(models.Server.name == server_spec['name']).first()
    if server is not None:
        # Check key
        key_ok = server.key != server_spec['key']
        if key_ok:
            logger.warning(u'Invalid server key supplied: name="{name}", key="{key}"'.format(**server_spec))
            raise Forbidden('Invalid server key')
    else:
        # Create
        server = models.Server(
            name=server_spec['name'],
            title=unicode(server_spec['name']),
            key=server_spec['key']
        )
        logger.info(u'Created new Server(name="{name}")'.format(**server_spec))

    # Update IP
    server.ip = request.remote_addr

    # Finish
    logger.debug(u'Identified server by name="{name}", id={id}'.format(id=server.id or '<new server>', **server_spec))
    return server


def _identify_service(ssn, server, service_name):
    """ Identify the service, create if missing
    :param ssn: Database session
    :type ssn: sqlalchemy.orm.session.Session
    :param server: Server to lookup the service at
    :type server: models.Server
    :param service_name: Service name on the server
    :type service_name: str
    :rtype: models.Service
    """
    service = ssn.query(models.Service).filter(
        models.Service.server == server,
        models.Service.name == service_name
    ).first()
    if service is None:
        service = models.Service(
            server=server,
            name=service_name,
            title=unicode(service_name)
        )
        logger.info(u'Created new Service(name="{name}", server="{server}")'.format(name=service.name, server=server.name))
    return service


@bp.route('/ping', methods=['POST'])
@jsonapi
def ping():
    """ Test connection and server credentials

        Status codes:
            400 invalid input
            403 invalid server key
    """
    ssn = g.db

    # Input validation
    data = request.get_json()
    assert isinstance(data, dict), 'Invalid data: should be JSON object'
    assert 'server' in data, 'Data: "server" key is missing'

    # Identify server (will raise exception if not fine)
    server = _identify_server(ssn, data['server'])
    ssn.add(server)
    ssn.commit()

    return {'pong': 1}


@bp.route('/set/service/status', methods=['POST'])
@jsonapi
def set_service_status():
    """ Receive single service status

        Status codes:
            400 invalid input
            403 invalid server key
    """
    ssn = g.db

    # Input validation
    data = request.get_json()
    assert isinstance(data, dict), 'Invalid data: should be JSON object'
    assert 'server' in data, 'Data: "server" key is missing'
    assert 'period' in data, 'Data: "period" key is missing'
    assert 'services' in data, 'Data: "services" key is missing'

    # Input validation: period
    try:
        data['period'] = int(data['period'])
    except ValueError:
        raise AssertionError('Data: "period" should be an integer')

    # Input validation: services
    assert isinstance(data['services'], list), 'Data: "services" should be a list'
    assert all(
        isinstance(s, dict) and
        'name' in s and isinstance(s['name'], basestring) and
        'state' in s and isinstance(s['state'], basestring) and
        ('info' not in s or isinstance(s['info'], basestring))
        for s in data['services']
    ), 'Data: "services" should be a list of objects with keys "name", "state", "info"?, "period"?'

    # Identify server
    server = _identify_server(ssn, data['server'])
    ssn.add(server)

    # Services
    services_cache = {}
    for s in data['services']:
        # Identify service
        try:
            service = services_cache[s['name']]
        except KeyError:
            service = _identify_service(ssn, server, s['name'])
            services_cache[service.name] = service
        ssn.add(service)

        # Update period
        try:
            service_period = int(s['period'])
        except (KeyError, ValueError):
            service_period = data['period']

        service.period = service_period

        # State
        if not models.state_t.is_valid(s['state']):
            s['info'] += u' (sent unsupported state: "{}")'.format(s['state'])
            s['state'] = 'UNK'
        state = models.ServiceState(
            service=service,
            rtime=datetime.utcnow(),
            state=s['state'],
            info=s['info']
        )
        ssn.add(state)
        logger.debug(u'Service {server}:`{name}` state update: {state}: {info}'.format(server=server.name, **s))

    # Save
    ssn.commit()

    return {'ok': 1}


@bp.route('/set/alerts', methods=['POST'])
@jsonapi
def set_alerts():
    """ Receive custom alerts

        Status codes:
            400 invalid input
            403 invalid server key
    """
    ssn = g.db

    # Input validation
    data = request.get_json()
    assert isinstance(data, dict), 'Invalid data: should be JSON object'
    assert 'server' in data, 'Data: "server" key is missing'
    assert 'alerts' in data, 'Data: "alerts" key is missing'

    # Input validation: alerts
    assert isinstance(data['alerts'], list), 'Data: "alerts" should be a list'
    assert all(
        isinstance(s, dict) and
        'message' in s and isinstance(s['message'], basestring) and
        ('service' not in s or isinstance(s['service'], basestring)) and
        set(s.keys()) <= {'title', 'message', 'service'}
        for s in data['alerts']
    ), 'Data: "alerts" should be a list of objects with keys "title", "message"?, "service"?'

    # Identify server
    server = _identify_server(ssn, data['server'])
    ssn.add(server)

    # Alerts
    services_cache = {}
    for a in data['alerts']:
        # Identify service (if any)
        service = None
        if 'service' in a:
            try:
                service = services_cache[a['service']]
            except KeyError:
                service = _identify_service(ssn, server, a['service'])
                services_cache[service.name] = service
                ssn.add(service)

        # Raise
        alert = models.Alert(
            server=server,
            service=service,
            ctime=datetime.utcnow(),
            channel='api',
            event='alert',
            message=unicode(a['message'])
        )
        ssn.add(alert)
        logger.debug(u'Alert reported for {server}:`{service}`: {message}'.format(server=server.name, service=service.name if service else '-', message=a['message']))

    # Save
    ssn.commit()

    return {'ok': 1}



#! /usr/bin/env python

import os
import logging
import multiprocessing

from overc import OvercApplication
from overc.lib.supervise import supervise_loop

def init_application(app_config_file):
    """ Initialize application from config file path
       :rtype: OvercApplication
       """
    # Load config
    app_config = OvercApplication.loadConfigFile(app_config_file)

    # Configure logging
    logging.basicConfig(level=getattr(logging, app_config['LOGLEVEL']))
    # TODO: multi-process app does not log everything fine! make it log to stdout, always

    # Applicaton
    application = OvercApplication(
        __name__,
        app_config['INSTANCE_PATH'],
        app_config
    )

    # Supervisor thread
    p = multiprocessing.Process(name='supervisor', target=supervise_loop, args=(application,))
    p.daemon = True
    p.start()

    # Finish
    return application

# Application
application = init_application(
    os.environ.get('OVERC_CONFIG', 'server.ini')
)

# WSGI entry point
app = application.app

# Serve static files
from werkzeug.wsgi import SharedDataMiddleware

app = SharedDataMiddleware(app, {
    '/ui/static/': ('overc.src.bps.ui', 'static')
}, cache=False)

# Debug mode
if __name__ == '__main__':
    from werkzeug.serving import run_simple
    run_simple(application=app,
               hostname='0.0.0.0',
               port=5000,
               use_reloader=True,
               use_debugger=True
    )
